from onegov.activity.models.occasion import Occasion
from onegov.activity.utils import random_group_code, dates_overlap
from onegov.core.orm import Base
from onegov.core.orm.mixins import TimestampMixin
from onegov.core.orm.types import UUID
from sqlalchemy import Column
from sqlalchemy import Enum
from sqlalchemy import ForeignKey
from sqlalchemy import func
from sqlalchemy import Index
from sqlalchemy import Integer
from sqlalchemy import Numeric
from sqlalchemy import Text
from sqlalchemy.ext.hybrid import hybrid_property
from sqlalchemy.orm import object_session, relationship
from sqlalchemy_utils import aggregated
from uuid import uuid4


class Booking(Base, TimestampMixin):
    """ Bookings are created by users for occasions.

    Initially, bookings are open. In this state they represent a "wish" rather
    than booking.

    Because bookings are wishes initially, they get a priority as well as
    a group code which links multiple bookings by multiple users together.

    They system will try to figure out the best booking using the priority
    as well as the group code.

    """

    __tablename__ = 'bookings'

    def __hash__(self):
        return hash(self.id)

    def __eq__(self, other):
        return self.id == other.id

    #: the public id of the booking
    id = Column(UUID, primary_key=True, default=uuid4)

    #: the user owning the booking
    username = Column(Text, ForeignKey('users.username'), nullable=False)

    #: the priority of the booking, a higher number = a higher priority
    priority = Column(Integer, nullable=False, default=0)

    #: the group code of the booking
    group_code = Column(Text, nullable=False, default=random_group_code)

    #: the attendee behind this booking
    attendee_id = Column(UUID, ForeignKey("attendees.id"), nullable=False)

    #: the occasion this booking belongs to
    occasion_id = Column(UUID, ForeignKey("occasions.id"), nullable=False)

    #: the cost of the booking
    cost = Column(Numeric(precision=8, scale=2), nullable=True)

    #: the period this booking belongs to
    @aggregated('occasion', Column(
        UUID, ForeignKey("periods.id"), nullable=False)
    )
    def period_id(self):
        return func.coalesce(Occasion.period_id, None)

    #: the state of the booking
    state = Column(
        Enum(
            'open',
            'blocked',
            'accepted',
            'denied',
            'cancelled',
            name='booking_state'
        ),
        nullable=False,
        default='open'
    )

    __mapper_args__ = {
        'order_by': priority
    }

    __table_args__ = (
        Index(
            'one_booking_per_attendee', 'occasion_id', 'attendee_id',
            unique=True
        ),
        Index('bookings_by_state', 'state', 'username')
    )

    #: access the user linked to this booking
    user = relationship('User')

    def period_bound_booking_state(self, period):
        """ During pre-booking we don't show the actual state of the booking,
        unless the occasion was cancelled, otherwise the user might see
        accepted bookings at a point where those states are not confirmed yet.

        This methods interprets the period/state accordingly.

        """
        if period.confirmed == True:
            return self.state

        return self.state == 'cancelled' and 'cancelled' or 'open'

    def provisional_booking_cost(self, period=None):
        """ The costs of the final booking, including the booking costs
        of the period (if not all-inclusive).

        This cost is written to the booking.cost when the period is
        confirmed.

        """

        period = period or self.period
        cost = self.occasion.cost or 0

        if not period.all_inclusive and period.booking_cost:
            cost += period.booking_cost

        return cost or 0

    def set_priority_bit(self, index, bit):
        """ Changes the priority, setting the the nth bit from the right to
        the value of ``bit`` (index/n begins at 0).

        The first bit (index=0) is reserved for starring/unstarring.
        The second bit (index=1) is reserved for nobble/unnobble.

        As a result, starring is less influental than locking.

        To give some context: Starring is used by the attendees to select
        which bookings they favor. Nobbling is used by administrators to force
        certain bookings to be preferred.

        """

        assert bit in (0, 1)
        assert index in (0, 1)

        mask = 1 << index

        self.priority &= ~mask

        if bit:
            self.priority |= mask

    def star(self, max_stars=3):
        """ Stars the current booking, up to a limit per period and attendee.

        Starring sets the star-bit to 1.

        :return: True if successful (or already set), False if over limit.

        """

        if self.starred:
            return True

        session = object_session(self)

        q = session.query(Booking.id)
        q = q.filter(Booking.attendee_id == self.attendee_id)
        q = q.filter(Booking.username == self.username)
        q = q.filter(Booking.period_id == self.period_id)
        q = q.filter(Booking.id != self.id)
        q = q.filter(Booking.starred == True)
        q = q.limit(max_stars + 1)

        if q.count() < max_stars:
            self.set_priority_bit(0, 1)
            return True

        return False

    def unstar(self):
        self.set_priority_bit(0, 0)

    def nobble(self):
        self.set_priority_bit(1, 1)

    def unnobble(self):
        self.set_priority_bit(1, 0)

    @hybrid_property
    def starred(self):
        return self.priority & 1 << 0 != 0

    @starred.expression
    def starred(self):
        return self.priority.op('&')(1 << 0) != 0

    @hybrid_property
    def nobbled(self):
        return self.priority & 1 << 1 != 0

    @nobbled.expression
    def nobbled(self):
        return self.priority.op('&')(1 << 1) != 0

    @property
    def dates(self):
        return self.occasion.dates

    @property
    def order(self):
        return self.occasion.order

    def overlaps(self, other):

        # even if exclude_from_overlap_check is active we consider a booking
        # to overlap itself (this protects against double bookings)
        if self.id == other.id:
            return True

        if self.occasion.exclude_from_overlap_check:
            return False

        # in some places 'other' only contains a start/end, so we can't
        # check for the overlap exclusion (those places are supposed to
        # do this on their own)
        if hasattr(other, 'occasion'):
            if other.occasion.exclude_from_overlap_check:
                return False

        return dates_overlap(
            tuple((d.start, d.end) for d in self.dates),
            tuple((o.start, o.end) for o in other.dates),
            minutes_between=self.period.minutes_between
        )



from onegov.activity.models import Activity, Attendee, Occasion
from onegov.user import User


class Scoring(object):
    """ Provides scoring based on a number of criteria.

    A criteria is a callable which takes a booking and returns a score.
    The final score is the sum of all criteria scores.

    """

    def __init__(self, criteria=None):
        self.criteria = criteria or [PreferMotivated()]

    def __call__(self, booking):
        return sum(criterium(booking) for criterium in self.criteria)

    @classmethod
    def from_settings(cls, settings, session):
        scoring = cls()

        if settings.get('prefer_in_age_bracket'):
            scoring.criteria.append(
                PreferInAgeBracket.from_session(session))

        if settings.get('prefer_organiser'):
            scoring.criteria.append(
                PreferOrganiserChildren.from_session(session))

        if settings.get('prefer_admins'):
            scoring.criteria.append(
                PreferAdminChildren.from_session(session))

        return scoring

    @property
    def settings(self):
        classes = {c.__class__ for c in self.criteria}
        settings = {}

        if PreferInAgeBracket in classes:
            settings['prefer_in_age_bracket'] = True

        if PreferOrganiserChildren in classes:
            settings['prefer_organiser'] = True

        if PreferAdminChildren in classes:
            settings['prefer_admins'] = True

        return settings


class PreferMotivated(object):
    """ Scores "motivated" bookings higher. A motivated booking is simply a
    booking with a higher priority (an attendee would favor a booking he's
    excited about.)

    """

    @classmethod
    def from_session(cls, session):
        return cls()

    def __call__(self, booking):
        return booking.priority


class PreferInAgeBracket(object):
    """ Scores bookings whose attendees fall into the age-bracket of the
    occasion higher.

    If the attendee falls into the age-bracket, the score is 1.0. Each year
    difference results in a penalty of 0.1, until 0.0 is reached.

    """

    def __init__(self, get_age_range, get_attendee_age):
        self.get_age_range = get_age_range
        self.get_attendee_age = get_attendee_age

    @classmethod
    def from_session(cls, session):
        attendees = None
        occasions = None

        def get_age_range(booking):
            nonlocal occasions, session

            if occasions is None:
                occasions = {
                    o.id: o.age
                    for o in session.query(Occasion.id, Occasion.age)
                    .filter(Occasion.period_id == booking.period_id)}

            return (
                occasions[booking.occasion_id].lower,
                occasions[booking.occasion_id].upper - 1
            )

        def get_attendee_age(booking):
            nonlocal attendees, session

            if attendees is None:
                attendees = {a.id: a.age for a in session.query(
                    Attendee.id, Attendee.age)}

            return attendees[booking.attendee_id]

        return cls(get_age_range, get_attendee_age)

    def __call__(self, booking):
        min_age, max_age = self.get_age_range(booking)
        attendee_age = self.get_attendee_age(booking)

        if min_age <= attendee_age and attendee_age <= max_age:
            return 1.0
        else:
            difference = min(
                abs(min_age - attendee_age),
                abs(max_age - attendee_age)
            )
            return 1.0 - min(1.0, difference / 10.0)


class PreferOrganiserChildren(object):
    """ Scores bookings of children higher if their parents are organisers.

    This is basically an incentive to become an organiser. A child whose parent
    is an organiser gets a score of 1.0, if the parent is not an organiser
    a score 0.0 is returned.

    """

    def __init__(self, get_is_organiser_child):
        self.get_is_organiser_child = get_is_organiser_child

    @classmethod
    def from_session(cls, session):
        organisers = None

        def get_is_organiser_child(booking):
            nonlocal organisers

            if organisers is None:
                organisers = {
                    a.username
                    for a in session.query(Activity.username)
                    .filter(Activity.id.in_(
                        session.query(Occasion.activity_id)
                        .filter(Occasion.period_id == booking.period_id)
                        .subquery()
                    ))
                }

            return booking.username in organisers

        return cls(get_is_organiser_child)

    def __call__(self, booking):
        return self.get_is_organiser_child(booking) and 1.0 or 0.0


class PreferAdminChildren(object):
    """ Scores bookings of children higher if their parents are admins. """

    def __init__(self, get_is_association_child):
        self.get_is_association_child = get_is_association_child

    @classmethod
    def from_session(cls, session):
        members = None

        def get_is_association_child(booking):
            nonlocal members

            if members is None:
                members = {
                    u.username for u in session.query(User)
                    .filter(User.role == 'admin')
                    .filter(User.active == True)
                }

            return booking.username in members

        return cls(get_is_association_child)

    def __call__(self, booking):
        return self.get_is_association_child(booking) and 1.0 or 0.0



import random

from datetime import date, timedelta
from freezegun import freeze_time
from functools import partial
from onegov.activity.matching import deferred_acceptance_from_database
from onegov.activity.matching import Scoring
from onegov.activity.matching import PreferAdminChildren
from onegov.activity.matching import PreferOrganiserChildren
from onegov.activity.matching import PreferInAgeBracket
from onegov.core.utils import Bunch
from psycopg2.extras import NumericRange
from uuid import uuid4


match = partial(
    deferred_acceptance_from_database,
    stability_check=True,
    validity_check=True,
    hard_budget=True
)


def new_occasion(collections, period, offset, length,
                 activity_name="foobar", username="owner@example.org",
                 spots=(0, 10), age=(0, 10)):

    activity = collections.activities.by_name(activity_name)

    if not activity:
        activity = collections.activities.add(activity_name, username=username)

    return collections.occasions.add(
        start=period.prebooking_start + timedelta(days=offset),
        end=period.prebooking_end + timedelta(days=offset + length),
        timezone="Europe/Zurich",
        activity=activity,
        period=period,
        spots=spots,
        age=age
    )


def new_attendee(collections, name=None, birth_date=None, user=None):

    return collections.attendees.add(
        name=name or uuid4().hex,
        birth_date=birth_date or date(2000, 1, 1),
        user=user or Bunch(username="owner@example.org"),
        gender=random.choice(('male', 'female')))


def test_simple_match(session, owner, collections, prebooking_period):
    """ Tests the matching of two occasions on the same activity with no
    conflicting signups.

    """

    o1 = new_occasion(collections, prebooking_period, 0, 1, spots=(0, 1))
    o2 = new_occasion(collections, prebooking_period, 0, 1, spots=(0, 1))

    a1 = new_attendee(collections)
    a2 = new_attendee(collections)

    collections.bookings.add(owner, a1, o1)
    collections.bookings.add(owner, a2, o2)

    # multiple runs lead to the same result
    for i in range(0, 2):
        match(session, prebooking_period.id)

        bookings = collections.bookings.query().all()

        assert len(bookings) == 2

        assert bookings[0].state == 'accepted'
        assert bookings[1].state == 'accepted'

        assert a1.happiness(prebooking_period.id) == 1.0
        assert a2.happiness(prebooking_period.id) == 1.0


def test_changing_priorities(session, owner, collections, prebooking_period):
    """ Makes sure that changing priorities lead to a changing match. """

    # only one available spot
    o = new_occasion(collections, prebooking_period, 0, 1, spots=(0, 1))

    # two attendees
    a1 = new_attendee(collections)
    a2 = new_attendee(collections)

    # one booking is prioritised
    b1 = collections.bookings.add(owner, a1, o)
    b2 = collections.bookings.add(owner, a2, o)

    b1.priority = 1
    b2.priority = 0

    match(session, prebooking_period.id)

    assert b1.state == 'accepted'
    assert b2.state == 'open'

    b1.priority = 0
    b2.priority = 1

    match(session, prebooking_period.id)

    assert b1.state == 'open'
    assert b2.state == 'accepted'


@freeze_time("2016-11-22")
def test_prefer_in_age_bracket(session, owner, collections, prebooking_period):
    o = new_occasion(
        collections, prebooking_period, 0, 1, spots=(0, 1), age=(10, 20))

    # two attendees, one is inside the age bracket, the other is outside
    a1 = new_attendee(collections, birth_date=date(1980, 11, 22))
    a2 = new_attendee(collections, birth_date=date(2000, 11, 22))

    b1 = collections.bookings.add(owner, a1, o)
    b2 = collections.bookings.add(owner, a2, o)

    match(session, prebooking_period.id, score_function=Scoring(
        criteria=[PreferInAgeBracket.from_session(session)]
    ))

    assert b1.state == 'open'
    assert b2.state == 'accepted'

    o.age = NumericRange(20, 50)

    match(session, prebooking_period.id, score_function=Scoring(
        criteria=[PreferInAgeBracket.from_session(session)]
    ))

    assert b1.state == 'accepted'
    assert b2.state == 'open'


def test_prefer_organisers_of_period(session, owner, secondary_owner,
                                     collections, prebooking_period,
                                     inactive_period):

    # if someone was an organiser in another period, it doesn't count
    new_occasion(collections, inactive_period, 0, 1, spots=(0, 1),
                 username=secondary_owner.username, activity_name="old")

    # what counts is being an organiser in the current period
    o = new_occasion(collections, prebooking_period, 0, 1, spots=(0, 1),
                     username=owner.username, activity_name="new")

    # two attendees, one belonging to an organiser
    a1 = new_attendee(collections, user=owner)
    a2 = new_attendee(collections, user=secondary_owner)

    b1 = collections.bookings.add(owner, a1, o)
    b2 = collections.bookings.add(secondary_owner, a2, o)

    match(session, prebooking_period.id, score_function=Scoring(
        criteria=[PreferOrganiserChildren.from_session(session)]
    ))

    assert b1.state == 'accepted'
    assert b2.state == 'open'


def test_prefer_organisers_over_members(session, owner, member,
                                        collections, prebooking_period):

    # organisers > members
    o = new_occasion(collections, prebooking_period, 0, 1, spots=(0, 1),
                     username=owner.username)

    # two attendees, one belonging to an organiser
    a1 = new_attendee(collections, user=member)
    a2 = new_attendee(collections, user=owner)

    b1 = collections.bookings.add(member, a1, o)
    b2 = collections.bookings.add(owner, a2, o)

    match(session, prebooking_period.id, score_function=Scoring(
        criteria=[PreferOrganiserChildren.from_session(session)]
    ))

    assert b1.state == 'open'
    assert b2.state == 'accepted'


def test_prefer_admin_children(session, owner, member, collections,
                               prebooking_period):

    owner.role = 'admin'

    o = new_occasion(collections, prebooking_period, 0, 1, spots=(0, 1))

    # two attendees, one belonging to an association member (editor)
    a1 = new_attendee(collections, user=member)
    a2 = new_attendee(collections, user=owner)

    b1 = collections.bookings.add(owner, a1, o)
    b2 = collections.bookings.add(member, a2, o)

    match(session, prebooking_period.id, score_function=Scoring(
        criteria=[PreferAdminChildren.from_session(session)]
    ))

    assert b1.state == 'accepted'
    assert b2.state == 'open'


def test_multiple_overlapping_dates(session, owner, collections,
                                    prebooking_period):

    o1 = new_occasion(collections, prebooking_period, 0, 1)
    collections.occasions.add_date(
        o1,
        o1.dates[0].end + timedelta(days=2, seconds=1),
        o1.dates[0].end + timedelta(days=3),
        o1.dates[0].timezone
    )

    o2 = new_occasion(collections, prebooking_period, 2, 1)
    collections.occasions.add_date(
        o2,
        o2.dates[0].end + timedelta(days=3, seconds=1),
        o2.dates[0].end + timedelta(days=4),
        o2.dates[0].timezone
    )

    a1 = new_attendee(collections, user=owner)

    b1 = collections.bookings.add(owner, a1, o1, priority=1)
    b2 = collections.bookings.add(owner, a1, o2, priority=0)

    match(session, prebooking_period.id)

    assert b1.state == 'accepted'
    assert b2.state == 'blocked'



""" Implements the matching algorithm used to match attendees to occasions.

The algorithm used is based on Deferred Acceptance. The algorithm has a
quadratic runtime.

"""

from onegov.activity import Attendee, Booking, Occasion, Period
from onegov.activity.matching.score import Scoring
from onegov.activity.matching.utils import overlaps, LoopBudget, hashable
from onegov.activity.matching.utils import booking_order, unblockable
from onegov.core.utils import Bunch
from itertools import groupby, product
from sortedcontainers import SortedSet
from sqlalchemy.orm import joinedload, defer


class AttendeeAgent(hashable('id')):
    """ Acts on behalf of the attendee with the goal to get a stable booking
    with an occasion.

    A booking/occasion pair is considered stable if there exists no other
    such pair which is preferred by both the attendee and the occasion.

    In other words, if there's no other occasion that would accept the
    attendee over another attendee.

    """

    __slots__ = ('id', 'wishlist', 'accepted', 'blocked')

    def __init__(self, id, bookings, limit=None, minutes_between=0):
        self.id = id
        self.limit = limit
        self.wishlist = SortedSet(bookings, key=booking_order)
        self.accepted = set()
        self.blocked = set()
        self.minutes_between = minutes_between

    def accept(self, booking):
        """ Accepts the given booking. """

        self.wishlist.remove(booking)
        self.accepted.add(booking)

        if self.limit and len(self.accepted) >= self.limit:
            self.blocked |= self.wishlist
        else:
            self.blocked |= {
                b for b in self.wishlist
                if overlaps(booking, b, self.minutes_between)
            }

        self.wishlist -= self.blocked

    def deny(self, booking):
        """ Removes the given booking from the accepted bookings. """

        self.wishlist.add(booking)
        self.accepted.remove(booking)

        # remove bookings from the blocked list which are not blocked anymore
        for booking in unblockable(self.accepted, self.blocked):
            if self.limit and len(self.wishlist) >= self.limit:
                break

            self.blocked.remove(booking)
            self.wishlist.add(booking)

    @property
    def is_valid(self):
        """ Returns True if the results of this agent are valid.

        The algorithm should never get to this stage, so this is an extra
        security measure to make sure there's no bug.

        """
        for a, b in product(self.accepted, self.accepted):
            if a != b and overlaps(a, b, self.minutes_between):
                return False

        return True


class OccasionAgent(hashable('id')):
    """ Represents the other side of the Attendee/Occasion pair.

    While the attende agent will try to get the best possible occasion
    according to the wishses of the attendee, the occasion agent will
    try to get the best attendee according to the wishes of the occasion.

    These wishes may include hard-coded rules or peferences defined by the
    organiser/admin, who may manually prefer certain attendees over others.

    """

    __slots__ = ('occasion', 'bookings', 'attendees', 'score_function')

    def __init__(self, occasion, score_function):
        self.id = occasion.id
        self.occasion = occasion
        self.bookings = set()
        self.attendees = {}
        self.score_function = score_function

    @property
    def full(self):
        return len(self.bookings) >= (self.occasion.max_spots)

    def preferred(self, booking):
        """ Returns the first booking with a lower score than the given
        booking (which indicates that the given booking is preferred over
        the returned item).

        If there's no preferred booking, None is returned.

        """
        return next(
            (
                b for b in self.bookings
                if self.score_function(b) < self.score_function(booking)
            ),
            None
        )

    def accept(self, attendee, booking):
        self.attendees[booking] = attendee
        self.bookings.add(booking)
        attendee.accept(booking)

    def deny(self, booking):
        self.attendees[booking].deny(booking)
        self.bookings.remove(booking)
        del self.attendees[booking]

    def match(self, attendee, booking):

        # as long as there are spots, automatically accept new requests
        if not self.full:
            self.accept(attendee, booking)

            return True

        # if the occasion is already full, accept the booking by throwing
        # another one out, if there exists a better fit
        over = self.preferred(booking)

        if over:
            self.deny(over)
            self.accept(attendee, booking)

            return True

        return False


def eager_score(bookings, score_function):
    """ Eagerly caches the booking scores by calculating them in advance.

    Returns a new score_function which uses those precomupted values.

    """
    scores = {b: score_function(b) for b in bookings}
    return scores.get


def deferred_acceptance_from_database(session, period_id, **kwargs):
    b = session.query(Booking)
    b = b.options(joinedload(Booking.occasion))
    b = b.filter(Booking.period_id == period_id)
    b = b.filter(Booking.state != 'cancelled')
    b = b.options(
        defer('group_code'),
    )

    o = session.query(Occasion)
    o = o.filter(Occasion.period_id == period_id)
    o = o.options(
        defer('meeting_point'),
        defer('note'),
        defer('cost')
    )

    period = session.query(Period).filter(Period.id == period_id).one()
    if period.all_inclusive and period.max_bookings_per_attendee:
        default_limit = period.max_bookings_per_attendee
        attendee_limits = None
    else:
        default_limit = None
        attendee_limits = {
            a.id: a.limit for a in
            session.query(Attendee.id, Attendee.limit)
        }

    bookings = deferred_acceptance(
        bookings=b, occasions=o,
        default_limit=default_limit, attendee_limits=attendee_limits,
        minutes_between=period.minutes_between, **kwargs)

    # write the changes to the database
    def update_states(bookings, state):
        ids = set(b.id for b in bookings)

        if not ids:
            return

        b = session.query(Booking)
        b = b.filter(Booking.id.in_(ids))

        for booking in b:
            booking.state = state

    update_states(bookings.open, 'open')
    update_states(bookings.accepted, 'accepted')
    update_states(bookings.blocked, 'blocked')


def deferred_acceptance(bookings, occasions,
                        score_function=None,
                        validity_check=True,
                        stability_check=False,
                        hard_budget=True,
                        default_limit=None,
                        attendee_limits=None,
                        minutes_between=0):
    """ Matches bookings with occasions.

    :score_function:
        A function accepting a booking and returning a score. Occasions prefer
        bookings with a higher score over bookings with a lower score, if and
        only if the occasion is not yet full.

        The score function is meant to return a constant value for each
        booking during the run of the algorithm. If this is not the case,
        the algorithm might not halt.

    :validity_check:
        Ensures that the algorithm doesn't lead to any overlapping bookings.
        Runs in O(b) time, where b is the number of bookings per period.

    :stability_check:
        Ensures that the result does not contain any blocking pairs, that is
        it checks that the result is stable. This runs in O(b^3) time, so
        do not run this in production (it's more of a testing tool).

    :hard_budget:
        Makes sure that the algorithm halts eventually by raising an exception
        if the runtime budget of O(a*b) is reached (number of attendees
        times the number of bookings).

        Feel free to proof that this can't happen and then remove the check ;)

    :default_limit:
        The maximum number of bookings which should be accepted for each
        attendee.

    :attendee_limits:
        The maximum number of bookings which should be accepted for each
        attendee. Keyed by the attendee id, this dictionary contains
        per-attendee limits. Those fall back to the default_limit.

    """
    attendee_limits = attendee_limits or {}
    score_function = eager_score(bookings, score_function or Scoring())

    bookings = [b for b in bookings]
    bookings.sort(key=lambda b: b.attendee_id)

    occasions = {
        o.id: OccasionAgent(o, score_function) for o in occasions
    }

    attendees = {
        aid: AttendeeAgent(
            aid,
            limit=attendee_limits.get(aid, default_limit),
            bookings=bookings,
            minutes_between=minutes_between
        )
        for aid, bookings in groupby(bookings, key=lambda b: b.attendee_id)
    }

    # I haven't proven yet that the following loop will always end. Until I
    # do there's a fallback check to make sure that we'll stop at some point
    budget = LoopBudget(max_ticks=len(bookings) * len(attendees))

    # while there are attendees with entries in a wishlist
    while next((a for a in attendees.values() if a.wishlist), None):

        if budget.limit_reached(as_exception=hard_budget):
            break

        candidates = [a for a in attendees.values() if a.wishlist]
        matched = 0

        # match attendees to courses
        while candidates:
            candidate = candidates.pop()

            for booking in candidate.wishlist:
                if occasions[booking.occasion_id].match(candidate, booking):
                    matched += 1
                    break  # required because the wishlist has been changed

        # if no matches were possible the situation can't be improved
        if not matched:
            break

    # make sure the algorithm didn't make any mistakes
    if validity_check:
        for a in attendees.values():
            assert a.is_valid

    # make sure the result is stable
    if stability_check:
        assert is_stable(attendees.values(), occasions.values())

    return Bunch(
        open=set(b for a in attendees.values() for b in a.wishlist),
        accepted=set(b for a in attendees.values() for b in a.accepted),
        blocked=set(b for a in attendees.values() for b in a.blocked)
    )


def is_stable(attendees, occasions):
    """ Returns true if the matching between attendees and occasions is
    stable.

    This runs in O(n^4) time, where n is the combination of
    bookings and occasions. So this is a testing tool, not something to
    run in production.

    """

    for attendee in attendees:
        for booking in attendee.accepted:
            for occasion in occasions:

                # the booking was actually accepted, skip
                if booking in occasion.bookings:
                    continue

                # if the current occasion prefers the given booking..
                over = occasion.preferred(booking)

                if over:
                    for o in occasions:
                        if o == occasion:
                            continue

                        # ..and another occasion prefers the loser..
                        switch = o.preferred(over)

                        # .. we have an unstable matching
                        if switch and occasion.preferred(switch):
                            return False

    return True



import logging
log = logging.getLogger('onegov.activity')  # noqa
log.addHandler(logging.NullHandler())  # noqa

from onegov.activity.models import (
    Activity,
    Attendee,
    Booking,
    InvoiceItem,
    Occasion,
    OccasionDate,
    Period
)
from onegov.activity.collections import (
    ActivityCollection,
    AttendeeCollection,
    BookingCollection,
    InvoiceItemCollection,
    OccasionCollection,
    PeriodCollection
)


__all__ = [
    'Activity',
    'Attendee',
    'Booking',
    'InvoiceItem',
    'Occasion',
    'OccasionDate',
    'Period',
    'ActivityCollection',
    'AttendeeCollection',
    'BookingCollection',
    'InvoiceItemCollection',
    'OccasionCollection',
    'PeriodCollection'
]



from onegov.activity import utils
from onegov.core.orm import Base
from onegov.core.orm.mixins import TimestampMixin
from onegov.core.orm.types import UUID
from onegov.user import User
from sqlalchemy import Boolean
from sqlalchemy import Column
from sqlalchemy import ForeignKey
from sqlalchemy import Numeric
from sqlalchemy import Text
from sqlalchemy.ext.hybrid import hybrid_property
from sqlalchemy.orm import validates, relationship
from sqlalchemy_utils import observes
from uuid import uuid4


class InvoiceItem(Base, TimestampMixin):
    """ An item in an invoice. """

    __tablename__ = 'invoice_items'

    #: the public id of the invoice item
    id = Column(UUID, primary_key=True, default=uuid4)

    #: the user who is charged
    username = Column(Text, ForeignKey('users.username'), nullable=False)
    user = relationship(User, backref="invoice_items")

    #: the invoice group (all items with the same text make one invoice)
    invoice = Column(Text, nullable=False)

    #: the code of the invoice used to identify the invoice through e-banking
    code = Column(Text, nullable=False)

    #: the item group (all items with the same text are visually grouped)
    group = Column(Text, nullable=False)

    #: the item text
    text = Column(Text, nullable=False)

    #: true if paid
    paid = Column(Boolean, nullable=False, default=False)

    #: the transaction id if paid through a bank or online transaction
    tid = Column(Text, nullable=True)

    #: the source of the transaction id, e.g. stripe, xml
    source = Column(Text, nullable=True)

    #: the unit to pay..
    unit = Column(Numeric(precision=8, scale=2), nullable=True)

    #: ..multiplied by the quantity..
    quantity = Column(Numeric(precision=8, scale=2), nullable=True)

    #: ..together form the amount
    @hybrid_property
    def amount(self):
        return self.unit * self.quantity

    @observes('invoice', 'username')
    def invoice_username_observer(self, invoice, username):
        self.code = utils.as_invoice_code(invoice, username)

    @validates('source')
    def validate_source(self, key, value):
        assert value in (None, 'xml', 'stripe')
        return value

    @property
    def display_code(self):
        """ The item's code, formatted in a human readable format. """
        return utils.format_invoice_code(self.code)

    @property
    def esr_code(self):
        """ The item's code, formatted as ESR. """
        return utils.encode_invoice_code(self.code)

    @property
    def display_esr_code(self):
        """ The item's ESR formatted in a human readable format. """
        return utils.format_esr_reference(self.esr_code)

    @property
    def discourage_changes(self):
        return self.source == 'xml'

    @property
    def disable_changes(self):
        return self.source and self.source != 'xml'



""" Contains upgrade tasks that are executed when the application is being
upgraded on the server. See :class:`onegov.core.upgrade.upgrade_task`.

"""
import hashlib

from onegov.activity import ActivityCollection
from onegov.activity import Booking
from onegov.activity import Period
from onegov.activity import Occasion
from onegov.activity import InvoiceItem
from onegov.core.orm.types import UUID, JSON
from onegov.core.upgrade import upgrade_task
from sqlalchemy import Boolean
from sqlalchemy import Column
from sqlalchemy import Date
from sqlalchemy import Enum
from sqlalchemy import ForeignKey
from sqlalchemy import Integer
from sqlalchemy import Numeric
from sqlalchemy import Text
from sqlalchemy.dialects.postgresql import ARRAY
from sqlalchemy.orm import joinedload


@upgrade_task('Rebuild bookings')
def rebuild_bookings(context):
    # having not created any bookings yet, we can rebuild them
    context.operations.drop_table('bookings')


@upgrade_task('Add period_id to bookings')
def add_period_id_to_bookings(context):

    context.operations.add_column('bookings', Column(
        'period_id', UUID, ForeignKey("periods.id"), nullable=True
    ))

    bookings = context.session.query(Booking)
    bookings = bookings.options(joinedload(Booking.occasion))

    for booking in bookings:
        booking.period_id = booking.occasion.period_id

    context.session.flush()
    context.operations.alter_column('bookings', 'period_id', nullable=False)


@upgrade_task('Change booking states')
def change_booking_states(context):

    new_type = Enum(
        'open',
        'blocked',
        'accepted',
        'denied',
        'cancelled',
        name='booking_state'
    )

    op = context.operations

    op.execute("""
        ALTER TABLE bookings ALTER COLUMN state TYPE Text;
        UPDATE bookings SET state = 'open' WHERE state = 'unconfirmed';
        DROP TYPE booking_state;
    """)

    new_type.create(op.get_bind())

    op.execute("""
        ALTER TABLE bookings ALTER COLUMN state
        TYPE booking_state USING state::text::booking_state;
    """)


@upgrade_task('Add confirmed flag to period')
def add_confirmed_flag_to_period(context):
    context.operations.add_column('periods', Column(
        'confirmed', Boolean, nullable=True, default=False
    ))

    for period in context.session.query(Period):
        period.confirmed = False

    context.session.flush()
    context.operations.alter_column('periods', 'confirmed', nullable=False)


@upgrade_task('Add data column to period')
def add_data_column_to_period(context):
    context.operations.add_column('periods', Column(
        'data', JSON, nullable=True, default=dict
    ))

    for period in context.session.query(Period):
        period.data = {}

    context.session.flush()
    context.operations.alter_column('periods', 'data', nullable=False)


@upgrade_task('Add attendee_count column to occasion')
def add_attendee_count_column_to_occasion(context):
    context.operations.add_column('occasions', Column(
        'attendee_count', Integer, nullable=True, default=0
    ))

    for occasion in context.session.query(Occasion):
        occasion.attendee_count = len(occasion.accepted)

    context.session.flush()
    context.operations.alter_column(
        'occasions', 'attendee_count', nullable=False)


@upgrade_task('Add finalized flag to period')
def add_finalized_flag_to_period(context):
    context.operations.add_column('periods', Column(
        'finalized', Boolean, nullable=True, default=False
    ))

    for period in context.session.query(Period):
        period.finalized = False

    context.session.flush()
    context.operations.alter_column('periods', 'finalized', nullable=False)


@upgrade_task('Add payment model columns')
def add_payment_model_columns(context):
    context.operations.add_column('periods', Column(
        'max_bookings_per_attendee', Integer, nullable=True
    ))

    context.operations.add_column('periods', Column(
        'booking_cost', Numeric(precision=8, scale=2), nullable=True
    ))

    context.operations.add_column('periods', Column(
        'all_inclusive', Boolean, nullable=True
    ))

    for period in context.session.query(Period):
        period.all_inclusive = False

    context.session.flush()
    context.operations.alter_column('periods', 'all_inclusive', nullable=False)

    context.operations.add_column('occasions', Column(
        'cost', Numeric(precision=8, scale=2), nullable=True
    ))

    context.operations.add_column('bookings', Column(
        'cost', Numeric(precision=8, scale=2), nullable=True
    ))


@upgrade_task('Add cancelled flag to occasion')
def add_cancelled_flag_to_occasion(context):
    context.operations.add_column('occasions', Column(
        'cancelled', Boolean, nullable=True
    ))

    for occasion in context.session.query(Occasion):
        occasion.cancelled = False

    context.session.flush()
    context.operations.alter_column('occasions', 'cancelled', nullable=False)


@upgrade_task('Add code field to invoice items')
def add_code_field_to_invoice_items(context):
    context.operations.add_column('invoice_items', Column(
        'code', Text, nullable=True
    ))

    for i in context.session.query(InvoiceItem):
        i.code = 'q' + ''.join((
            hashlib.sha1((
                i.invoice + i.username).encode('utf-8')
            ).hexdigest()[:5],
            hashlib.sha1(
                i.username.encode('utf-8')
            ).hexdigest()[:5]
        ))

    context.session.flush()
    context.operations.alter_column('invoice_items', 'code', nullable=False)


@upgrade_task('Add source field to invoice items')
def add_source_field_to_invoice_items(context):
    if not context.has_column('invoice_items', 'source'):
        context.operations.add_column('invoice_items', Column(
            'source', Text, nullable=True
        ))


@upgrade_task('Add gender/notes fields to attendees')
def add_gender_notes_fields_to_attendees(context):
    if not context.has_column('attendees', 'gender'):
        context.operations.add_column('attendees', Column(
            'gender', Text, nullable=True
        ))

    if not context.has_column('attendees', 'notes'):
        context.operations.add_column('attendees', Column(
            'notes', Text, nullable=True
        ))


@upgrade_task('Support multiple dates per occasion')
def support_multiple_dates_per_occasion(context):
    context.operations.drop_constraint('start_before_end', 'occasions')

    for name in ('durations', 'order'):
        context.operations.add_column(
            'occasions', Column(name, Integer, server_default="0"))
        context.operations.alter_column(
            'occasions', name, server_default=None)

    context.session.execute("""
        INSERT INTO occasion_dates ("timezone", "start", "end", "occasion_id")
        SELECT "timezone", "start", "end", "id" FROM occasions
    """)

    context.session.flush()

    # update dates
    for occasion in context.session.query(Occasion):
        occasion.on_date_change()

    for name in ('start', 'end', 'timezone'):
        context.operations.drop_column('occasions', name)


@upgrade_task('Adds minutes_between to period')
def adds_minutes_between_to_period(context):
    context.operations.add_column('periods', Column(
        'minutes_between', Integer, nullable=True, server_default='0'))
    context.operations.alter_column(
        'periods', 'minutes_between', server_default=None)


@upgrade_task('Adds exclude_from_overlap_check to period')
def adds_exclude_from_overlap_check_to_period(context):
    context.operations.add_column('occasions', Column(
        'exclude_from_overlap_check', Boolean, nullable=False,
        server_default='FALSE'))
    context.operations.alter_column(
        'occasions', 'exclude_from_overlap_check', server_default=None)


@upgrade_task('Adds deadlines to period')
def adds_deadlines_to_period(context):
    if not context.has_column('periods', 'deadline_date'):
        context.operations.add_column('periods', Column(
            'deadline_date', Date, nullable=True
        ))
    if not context.has_column('periods', 'deadline_days'):
        context.operations.add_column('periods', Column(
            'deadline_days', Integer, nullable=True
        ))


@upgrade_task('Adds limit to attendee')
def adds_limit_to_attendee(context):
    if not context.has_column('attendees', 'limit'):
        context.operations.add_column('attendees', Column(
            'limit', Integer, nullable=True
        ))


@upgrade_task('Introduce location/meeting_point')
def introduce_location_meeting_point(context):
    if not context.has_column('activities', 'location'):
        context.operations.add_column('activities', Column(
            'location', Text, nullable=True
        ))
    if not context.has_column('occasions', 'meeting_point'):
        context.operations.alter_column(
            'occasions', 'location', new_column_name='meeting_point')


@upgrade_task('Add active days')
def add_active_days(context):
    if context.has_column('activities', 'active_days'):
        assert context.has_column('occasions', 'active_days')
        return

    context.session.execute(
        'CREATE AGGREGATE "{}".array_cat_agg(anyarray) '
        '(SFUNC=array_cat, STYPE=anyarray)'.format(
            context.schema
        )
    )

    context.operations.add_column('activities', Column(
        'active_days', ARRAY(Integer), nullable=True
    ))

    context.operations.add_column('occasions', Column(
        'active_days', ARRAY(Integer), nullable=True
    ))

    context.session.flush()

    # update dates
    for occasion in context.session.query(Occasion):
        occasion.on_date_change()


@upgrade_task('Add active days index')
def add_active_days_index(context):
    context.operations.create_index(
        'inverted_active_days', 'activities', ['active_days'],
        postgresql_using='gin'
    )


@upgrade_task('Removed denied activity state')
def remove_denied_activity_state(context):

    new_type = Enum(
        'preview',
        'proposed',
        'accepted',
        'archived',
        name='activity_state'
    )

    op = context.operations

    op.execute("""
        ALTER TABLE activities ALTER COLUMN state TYPE Text;
        UPDATE activities SET state = 'archived' WHERE state = 'denied';
        DROP TYPE activity_state;
    """)

    new_type.create(op.get_bind())

    op.execute("""
        ALTER TABLE activities ALTER COLUMN state
        TYPE activity_state USING state::text::activity_state;
    """)


@upgrade_task('Add weekdays')
def add_weekdays(context):
    if context.has_column('activities', 'weekdays'):
        assert context.has_column('occasions', 'weekdays')
        return

    context.operations.add_column('activities', Column(
        'weekdays', ARRAY(Integer), nullable=True
    ))

    context.operations.add_column('occasions', Column(
        'weekdays', ARRAY(Integer), nullable=True
    ))

    context.session.flush()

    # update dates
    for occasion in context.session.query(Occasion):
        occasion.on_date_change()


@upgrade_task('Add weekdays index')
def add_weekdays_index(context):
    context.operations.create_index(
        'inverted_weekdays', 'activities', ['weekdays'],
        postgresql_using='gin'
    )


@upgrade_task('Extract thumbnails')
def extract_thumbnails(context):
    for activity in ActivityCollection(context.session).query():
        activity.content_observer(None)



import sedate

from enum import IntEnum
from onegov.core.orm import Base
from onegov.core.orm.mixins import TimestampMixin
from onegov.core.orm.types import UUID, UTCDateTime
from sqlalchemy import case
from sqlalchemy import event
from sqlalchemy import func
from sqlalchemy import CheckConstraint
from sqlalchemy import Column
from sqlalchemy import ForeignKey
from sqlalchemy import Integer
from sqlalchemy import Text
from sqlalchemy.ext.hybrid import hybrid_property
from sqlalchemy.orm import Session


class DAYS(IntEnum):
    half = 2**1
    full = 2**2
    many = 2**3

    @staticmethod
    def has(value, mask):
        return value & mask > 0 if value else False


class OccasionDate(Base, TimestampMixin):
    """ A single start/end date of an occurrence (which may have multiple
    date ranges).

    """

    __tablename__ = 'occasion_dates'

    def __hash__(self):
        return hash(self.id)

    #: the internal id of this occasion date
    id = Column(Integer, primary_key=True)

    #: Timezone of the occasion date
    timezone = Column(Text, nullable=False)

    #: The start of the range
    start = Column(UTCDateTime, nullable=False)

    #: The end of the range
    end = Column(UTCDateTime, nullable=False)

    #: The associated occasion
    occasion_id = Column(UUID, ForeignKey('occasions.id'), nullable=False)

    __table_args__ = (
        CheckConstraint('"start" <= "end"', name='start_before_end'),
    )

    @property
    def localized_start(self):
        return sedate.to_timezone(self.start, self.timezone)

    @property
    def localized_end(self):
        return sedate.to_timezone(self.end, self.timezone)

    @property
    def active_days(self):
        for dt in sedate.dtrange(self.localized_start, self.localized_end):
            yield dt.date().toordinal()

    @property
    def weekdays(self):
        return list({
            dt.weekday() for dt in sedate.dtrange(
                self.localized_start, self.localized_end
            )
        })

    @hybrid_property
    def duration_in_seconds(self):
        return (self.end - self.start).total_seconds()

    @duration_in_seconds.expression
    def duration_in_seconds(self):
        return func.extract('epoch', self.end - self.start)

    @hybrid_property
    def duration(self):
        hours = self.duration_in_seconds / 3600

        # defined here and in the expression below!
        if hours <= 6:
            return DAYS.half
        elif hours <= 24:
            return DAYS.full
        else:
            return DAYS.many

    @duration.expression
    def duration(self):

        # defined here and in the property above!
        return case((
            (OccasionDate.duration_in_seconds <= (6 * 3600), int(DAYS.half)),
            (OccasionDate.duration_in_seconds <= (24 * 3600), int(DAYS.full)),
        ), else_=int(DAYS.many))

    def overlaps(self, other):
        return sedate.overlaps(self.start, self.end, other.start, other.end)


# # changes to the dates need to be propagated to the parent occasion
# # so it can update its aggreagated values
@event.listens_for(Session, 'before_flush')
def before_flush(session, context, instances):
    for obj in session.dirty:
        if isinstance(obj, OccasionDate):
            obj.occasion.on_date_change()



from onegov.activity.models.activity import Activity, ACTIVITY_STATES
from onegov.activity.models.attendee import Attendee
from onegov.activity.models.booking import Booking
from onegov.activity.models.invoice_item import InvoiceItem
from onegov.activity.models.occasion import Occasion
from onegov.activity.models.occasion_date import OccasionDate, DAYS
from onegov.activity.models.period import Period


__all__ = [
    'Activity',
    'Attendee',
    'Booking',
    'InvoiceItem',
    'Occasion',
    'OccasionDate',
    'Period',
    'ACTIVITY_STATES',
    'DAYS'
]



from onegov.activity.models import InvoiceItem
from onegov.core.collection import GenericCollection
from sqlalchemy import func, not_


class InvoiceItemCollection(GenericCollection):

    def __init__(self, session, username=None, invoice=None):
        super().__init__(session)
        self.username = username
        self.invoice = invoice

    def query(self):
        query = super().query()

        if self.username is not None:
            query = query.filter(self.model_class.username == self.username)

        if self.invoice is not None:
            query = query.filter(self.model_class.invoice == self.invoice)

        return query

    def for_username(self, username):
        return self.__class__(self.session, username, self.invoice)

    def for_invoice(self, invoice):
        return self.__class__(self.session, self.username, invoice)

    @property
    def model_class(self):
        return InvoiceItem

    @property
    def total(self):
        q = self.query()
        q = q.with_entities(func.sum(InvoiceItem.amount))

        return q.scalar()

    @property
    def outstanding(self):
        q = self.query()
        q = q.filter(InvoiceItem.paid == False)
        q = q.with_entities(func.sum(InvoiceItem.amount))

        return q.scalar()

    def count_unpaid_invoices(self, exclude_invoices=None):
        q = self.query()
        q = q.with_entities(func.count(func.distinct(InvoiceItem.invoice)))
        q = q.filter(InvoiceItem.paid == False)

        if exclude_invoices:
            q = q.filter(not_(InvoiceItem.invoice.in_(exclude_invoices)))

        return q.scalar()

    def add(self, user, invoice, group, text, unit, quantity):
        if isinstance(user, str):
            username = user
        else:
            username = user.username

        return super().add(
            username=username,
            invoice=invoice,
            group=group,
            text=text,
            unit=unit,
            quantity=quantity
        )



import pytest

from onegov.activity.utils import append_checksum
from onegov.activity.utils import as_invoice_code
from onegov.activity.utils import decode_esr_reference
from onegov.activity.utils import encode_invoice_code
from onegov.activity.utils import format_esr_reference
from onegov.activity.utils import format_invoice_code
from onegov.activity.utils import generate_checksum
from onegov.activity.utils import is_valid_checksum
from onegov.activity.utils import merge_ranges


def test_merge_ranges():
    assert merge_ranges([[1, 2]]) == [(1, 2)]
    assert merge_ranges([[1, 2], [2, 3]]) == [(1, 3)]
    assert merge_ranges([[1, 2], [2, 3]]) == [(1, 3)]
    assert merge_ranges([[1, 2], [3, 4]]) == [(1, 2), (3, 4)]
    assert merge_ranges([[1, 2], [1, 2], [2, 3]]) == [(1, 3)]


def test_generate_checksum():
    assert generate_checksum("96111690000000660000000928") == 4
    assert generate_checksum("12000000000023447894321689") == 9


def test_append_checksum():
    assert append_checksum("96111690000000660000000928")\
        == "961116900000006600000009284"
    assert append_checksum("12000000000023447894321689")\
        == "120000000000234478943216899"


def test_is_valid_checksum():
    assert is_valid_checksum("961116900000006600000009284") is True
    assert is_valid_checksum("120000000000234478943216899") is True
    assert is_valid_checksum("961116900000006600000009285") is False
    assert is_valid_checksum("120000000000234478943216898") is False


def test_as_invoice_code():
    assert as_invoice_code('first', 'foo@example.org') == 'qeb3afd0e43'
    assert as_invoice_code('second', 'foo@example.org') == 'q9df17d0e43'
    assert as_invoice_code('first', 'bar@example.org') == 'q3717d883ed'


def test_format_invoice_code():
    assert format_invoice_code('qeb3afd0e43') == 'Q-EB3AF-D0E43'


def test_encode_invoice_code():
    assert encode_invoice_code('qca7df00e15') == '000127131108141601011502061'
    assert is_valid_checksum('000127131108141601011502061')


def test_decode_esr_reference():
    assert decode_esr_reference('000127131108141601011502061') == 'qca7df00e15'
    assert decode_esr_reference('127131108141601011502061') == 'qca7df00e15'
    assert decode_esr_reference('1 2713 1108141601011502061') == 'qca7df00e15'

    def assert_raises_error(code, message):
        with pytest.raises(RuntimeError) as e:
            decode_esr_reference(code)
        assert e.value.args[0] == message

    assert_raises_error(
        '000227131108141601011502069',
        'Unknown ESR reference version: 2')

    assert_raises_error(
        '000227131108141601011502069123123',
        'ESR reference is too long')

    assert_raises_error(
        '127131108141601011502062',
        'ESR reference has an invalid checksum')

    assert_raises_error(
        '11',
        'ESR reference is too short'
    )

    assert_raises_error(
        '1271311081416010115020610',
        'ESR reference has an uneven number of digits'
    )


def test_format_esr_reference():
    assert format_esr_reference('000127131108141601011502061')\
        == '1271 31108 14160 10115 02061'

    assert format_esr_reference('127131108141601011502061')\
        == '1271 31108 14160 10115 02061'



import hashlib
import lxml
import sedate
import string
import re

from datetime import timedelta
from itertools import chain
from onegov.core.utils import chunks
from pyquery import PyQuery as pq
from random import SystemRandom


GROUP_CODE_CHARS = string.ascii_uppercase + string.digits
GROUP_CODE_LENGTH = 10

CODE_TO_ESR_MAPPING = {
    character: '{:02d}'.format(value) for value, character in chain(
        enumerate(string.digits, start=1),
        enumerate(string.ascii_lowercase, start=11)
    )
}

ESR_TO_CODE_MAPPING = {
    value: key for key, value in CODE_TO_ESR_MAPPING.items()
}

INTERNAL_IMAGE_EX = re.compile(r'.*/storage/[0-9a-z]{64}')


def random_group_code():
    random = SystemRandom()

    return ''.join(
        random.choice(GROUP_CODE_CHARS)
        for _ in range(GROUP_CODE_LENGTH)
    )


def overlaps(range_a, range_b):
    return range_b[0] <= range_a[0] and range_a[0] <= range_b[1] or\
        range_a[0] <= range_b[0] and range_b[0] <= range_a[1]


def merge_ranges(ranges):
    """ Merges the given list of ranges into a list of ranges including only
    exclusive ranges. The ranges are turned into tuples to make them
    hashable.

    """

    ranges = sorted(list(ranges))

    # stack of merged values
    merged = [tuple(ranges[0])]

    for r in ranges:
        if overlaps(merged[-1], r):
            merged[-1] = (merged[-1][0], r[1])
        else:
            merged.append(tuple(r))

    return merged


def generate_checksum(number):
    """ Generates the modulo 10 checksum as required by Postfinance.

    :return: The checksum as integer.

    """

    table = (0, 9, 4, 6, 8, 2, 7, 1, 3, 5)
    carry = 0

    for n in str(number):
        carry = table[(carry + int(n)) % 10]

    return (10 - carry) % 10


def append_checksum(number):
    """ Adds the Postfinance modulo 10 checksum to the given number.

    :return: The given number as string with the checksum appended.

    """
    number = str(number)
    return number + str(generate_checksum(number))


def is_valid_checksum(number):
    """ Takes the given number and verifys the checksum which is assumed to
    be the last digit in the given number.

    :return: True if the checksum digit matches the numbers before it.

    """

    number = str(number)
    return number[-1] == str(generate_checksum(number[:-1]))


def as_invoice_code(invoice, username):
    """ Takes the invoice group and username from an
    :class:`~onegov.activity.models.invoice_item.InvoiceItem` and generates
    a hash unique to this invoice/username combination.

    There's no guarantee that this code is unique for an invoice, though
    the chance of it overlapping is very very small -> any algorithm
    doing some kind of matching has to account for this fact.

    We can solve this by introducing a separate invoice record in the future.

    The "q" at the beginning is used as a marker for regular expressions
    (so it's a "q", followed by [a-z0-9]{8}), as well as a version
    identifier.

    If this code changes be aware that the esr encoding/decoding might have
    to be changed as well.

    """

    return 'q' + ''.join((
        hashlib.sha1((invoice + username).encode('utf-8')).hexdigest()[:5],
        hashlib.sha1(username.encode('utf-8')).hexdigest()[:5]
    ))


def format_invoice_code(code):
    """ Takes the invoice code and formats it in a human readable way. """

    code = code.upper()

    return '-'.join((
        code[:1],
        code[1:6],
        code[6:]
    ))


def encode_invoice_code(code):
    """ Takes an invoice code and encodes it as a valid ESR reference with
    checksum included.

    """

    # we prepend a version number in front of the code which goes from
    # 1-9, so we have that many tries to get it right ;)
    version = '1'

    blocks = [version]

    for char in code:
        if char in CODE_TO_ESR_MAPPING:
            blocks.append(CODE_TO_ESR_MAPPING[char])
        else:
            raise RuntimeError("Invalid character {} in {}".format(char, code))

    return append_checksum('{:0>26}'.format(''.join(blocks)))


def decode_esr_reference(reference):
    """ Takes an ESR reference and decodes it into an invoice code if possible.

    Raises an error if the given reference cannot be decoded.

    Handles formatted refences as well as unformatted ones.

    """
    reference = reference.replace(' ', '')

    if len(reference) > 27:
        raise RuntimeError("ESR reference is too long")

    if not is_valid_checksum(reference):
        raise RuntimeError("ESR reference has an invalid checksum")

    reference = reference.lstrip('0')
    version, encoded = reference[0], reference[1:-1]

    if version != '1':
        raise RuntimeError("Unknown ESR reference version: {}".format(version))

    if len(reference) < 22:
        raise RuntimeError("ESR reference is too short")

    if len(reference) % 2 != 0:
        raise RuntimeError("ESR reference has an uneven number of digits")

    blocks = []

    for items in chunks(encoded, 2):
        char = ''.join(items)

        # at this point all characters are going to be valid
        # by virtue of our checksum check
        blocks.append(ESR_TO_CODE_MAPPING[char])

    return ''.join(blocks)


def format_esr_reference(reference):
    """ Takes an ESR reference and formats it in a human-readable way. """
    reference = reference.lstrip('0')
    reference = reference.replace(' ', '')

    blocks = []

    for values in chunks(reversed(reference), n=5, fillvalue=''):
        blocks.append(''.join(reversed(values)))

    return ' '.join(reversed(blocks))


def generate_xml(payments):
    """ Creates an xml for import through ISO20022. Used for testing only. """

    transactions = []

    default = {
        'reference': '',
        'note': ''
    }

    for ix, payment in enumerate(payments):

        if 'tid' not in payment:
            payment['tid'] = 'T{}'.format(ix)

        if payment['amount'].startswith('-'):
            payment['credit'] = 'DBIT'
        else:
            payment['credit'] = 'CRDT'

        payment['currency'] = payment['amount'][-3:]
        payment['amount'] = payment['amount'].strip('-+')[:-3]

        for key in default:
            if key not in payment:
                payment[key] = default[key]

        transactions.append("""
        <TxDtls>
            <Refs>
                <AcctSvcrRef>{tid}</AcctSvcrRef>
            </Refs>
            <Amt Ccy="{currency}">{amount}</Amt>
            <CdtDbtInd>{credit}</CdtDbtInd>
            <RmtInf>
                <Strd>
                    <CdtrRefInf>
                        <Ref>{reference}</Ref>
                    </CdtrRefInf>
                </Strd>
                <Ustrd>{note}</Ustrd>
            </RmtInf>
        </TxDtls>
        """.format(**payment))

    return """<?xml version="1.0" encoding="UTF-8"?>
        <Document>
            <BkToCstmrStmt>
                <Stmt>
                    <Ntry>
                        <NtryDtls>
                            {}
                        </NtryDtls>
                    </Ntry>
                </Stmt>
            </BkToCstmrStmt>
        </Document>
    """.format('\n'.join(transactions))


def dates_overlap(a, b, minutes_between=0, cut_end=True):
    """ Returns true if any time tuple in the list of tuples in a overlaps
    with a time tuple in b.

    """
    # this can be done with an O(n log n) algorithm but since we are
    # operating on a very small n the constant factors dominate and there
    # are fewer constant factors in this approach:

    offset = timedelta(seconds=minutes_between / 2 * 60)
    ms = cut_end and timedelta(microseconds=1) or timedelta()

    # make sure that 11:00 - 12:00 and 12:00 - 13:00 are not overlapping
    ms = timedelta(microseconds=1)

    for s, e in a:
        for os, oe in b:
            if sedate.overlaps(
                s - offset, e + offset - ms,
                os - offset, oe + offset - ms
            ):
                return True

    return False


def is_internal_image(url):
    return url and INTERNAL_IMAGE_EX.match(url) and True or False


def extract_thumbnail(text):

    try:
        first_image = next((img for img in pq(text or '')('img')), None)
    except lxml.etree.XMLSyntaxError:
        first_image = None

    if first_image is None:
        return None

    url = first_image.attrib.get('src')

    if is_internal_image(url) and not url.endswith('/thumbnail'):
        url += '/thumbnail'

    return url



from onegov.activity import log
from onegov.activity.utils import dates_overlap
from sortedcontainers import SortedSet


def overlaps(booking, other, minutes_between=0):
    """ Returns true if the two given bookings overlap. """

    assert booking != other

    if booking.occasion.exclude_from_overlap_check:
        return False

    if other.occasion.exclude_from_overlap_check:
        return False

    return dates_overlap(
        tuple((b.start, b.end) for b in booking.dates),
        tuple((o.start, o.end) for o in other.dates),
        minutes_between=minutes_between
    )


class LoopBudget(object):
    """ Helps ensure that a loop doesn't overreach its complexity budget.

    For example::

        budget = LoopBudget(max_ticks=10)

        while True:
            if budget.limit_reached():
                break
    """

    def __init__(self, max_ticks):
        self.ticks = 0
        self.max_ticks = max_ticks

    def limit_reached(self, as_exception=False):
        if self.ticks >= self.max_ticks:
            message = "Loop limit of {} has been reached".format(self.ticks)

            if as_exception:
                raise RuntimeError(message)
            else:
                log.warning(message)

            return True

        self.ticks += 1


def hashable(attribute):

    class Hashable(object):

        def __hash__(self):
            return hash(getattr(self, attribute))

        def __eq__(self, other):
            return getattr(self, attribute) == getattr(other, attribute)

    return Hashable


def booking_order(booking):
    """ Keeps the bookings predictably sorted from highest to lowest priority.

    """

    return booking.priority * -1, booking.id


def unblockable(accepted, blocked, key=booking_order):
    """ Returns a set of items in the blocked set which do not block
    with anything. The set is ordered using :func:`booking_order`.

    """

    unblockable = SortedSet(blocked, key=key)

    for accepted in accepted:
        for booking in blocked:
            if accepted.overlaps(booking):
                unblockable.remove(booking)

    return unblockable



from onegov.activity.matching.core import deferred_acceptance
from onegov.activity.matching.core import deferred_acceptance_from_database
from onegov.activity.matching.interfaces import MatchableBooking
from onegov.activity.matching.interfaces import MatchableOccasion
from onegov.activity.matching.score import Scoring
from onegov.activity.matching.score import PreferMotivated
from onegov.activity.matching.score import PreferInAgeBracket
from onegov.activity.matching.score import PreferOrganiserChildren
from onegov.activity.matching.score import PreferAdminChildren

__all__ = [
    'deferred_acceptance',
    'deferred_acceptance_from_database',
    'MatchableBooking',
    'MatchableOccasion',
    'Scoring',
    'PreferMotivated',
    'PreferInAgeBracket',
    'PreferOrganiserChildren',
    'PreferAdminChildren',
]



import re

from collections import defaultdict, namedtuple, Counter
from datetime import date
from decimal import Decimal
from itertools import groupby
from lxml import etree
from onegov.activity.models import InvoiceItem
from onegov.activity.utils import encode_invoice_code
from pprint import pformat


DOCUMENT_NS_EX = re.compile(r'.*<Document [^>]+>(.*)')
INVALID_CODE_CHARS_EX = re.compile(r'[^Q0-9A-F]+')
CODE_EX = re.compile(r'Q{1}[A-F0-9]{10}')


def normalize_xml(xml):
    # let's not bother with namespaces at all
    return DOCUMENT_NS_EX.sub(r'<Document>\1', xml)


class Transaction(object):

    __slots__ = (
        'amount',
        'booking_date',
        'booking_text',
        'confidence',
        'credit',
        'currency',
        'debitor',
        'debitor_account',
        'duplicate',
        'note',
        'paid',
        'reference',
        'tid',
        'username',
        'valuta_date',
    )

    def __init__(self, **kwargs):
        for key, value in kwargs.items():
            setattr(self, key, value)

        self.username = None
        self.confidence = 0
        self.duplicate = False
        self.paid = False

    def __repr__(self):
        return pformat({key: getattr(self, key) for key in self.__slots__})

    @property
    def code(self):
        return extract_code(self.booking_text) or extract_code(self.note)

    @property
    def order(self):
        state = self.state

        if self.valuta_date:
            date = self.valuta_date and (
                self.valuta_date.year * 10000 +
                self.valuta_date.month * 100 +
                self.valuta_date.day
            ) * - 1
        else:
            date = -1

        if state == 'success':
            return 0, date, self.username

        if state == 'paid':
            return 1, date, self.username

        if state == 'duplicate':
            return 2, date, self.username

        if state == 'warning':
            return 3, date, self.username

        return 4, date

    @property
    def state(self):
        if self.paid:
            return 'paid'

        if self.duplicate:
            return 'duplicate'

        if self.confidence == 1:
            return 'success'

        if self.confidence == 0.5:
            return 'warning'

        return 'unknown'


def extract_transactions(xml):
    root = etree.fromstring(normalize_xml(xml).encode('utf-8'))

    def first(element, xpath):
        elements = element.xpath(xpath)
        return elements[0] if elements else None

    def joined(element, xpath):
        return '\n'.join(element.xpath(xpath))

    def as_decimal(text):
        if text:
            return Decimal(text)

    def as_date(text):
        if text:
            return date(*[int(p) for p in text.split('-')])

    for entry in root.xpath('/Document/BkToCstmrStmt/Stmt/Ntry'):
        booking_date = as_date(first(entry, 'BookgDt/Dt/text()'))
        valuta_date = as_date(first(entry, 'ValDt/Dt/text()'))
        booking_text = first(entry, 'AddtlNtryInf/text()')

        for d in entry.xpath('NtryDtls/TxDtls'):
            yield Transaction(
                booking_date=booking_date,
                valuta_date=valuta_date,
                booking_text=booking_text,
                tid=first(d, 'Refs/AcctSvcrRef/text()'),
                amount=as_decimal(first(d, 'Amt/text()')),
                currency=first(d, 'Amt/@Ccy'),
                reference=first(d, 'RmtInf/Strd/CdtrRefInf/Ref/text()'),
                note=joined(d, 'RmtInf/Ustrd/text()'),
                credit=first(d, 'CdtDbtInd/text()') == 'CRDT',
                debitor=first(d, 'RltdPties/Dbtr/Nm/text()'),
                debitor_account=first(d, 'RltdPties/DbtrAcct/Id/IBAN/text()'),
            )


def extract_code(text):
    """ Takes a bunch of text and tries to extract a code from it.

    :return: The code without formatting and in lowercase or None

    """

    if text is None:
        return None

    text = text.replace('\n', '').strip()

    if not text:
        return None

    # ENTER A WORLD WITHOUT LOWERCASE
    text = text.upper()

    # replace all O-s (as in OMG) with 0.
    text = text.replace('O', '0')

    # normalize the text by removing all invalid characters.
    text = INVALID_CODE_CHARS_EX.sub('', text)

    # try to fetch the code
    match = CODE_EX.search(text)

    if match:
        return match.group().lower()
    else:
        return None


def match_camt_053_to_usernames(xml, collection, invoice, currency='CHF'):
    """ Takes an ISO20022 camt.053 file and matches it with the invoice
    items in the :class:`~onegov.activity.collections.InvoiceItemCollection`.

    Raises an error if the given xml cannot be processed.

    :return: A list of transactions found in the xml file, together with
    the matching username and a confidence attribute indicating how
    certain the match is (1.0 indicating a sure match, 0.5 a possible match
    and 0.0 a non-match).

    """

    # Get all paid transaction ids to be mark transactions which were paid
    q = collection.for_invoice(None).query()
    q = q.with_entities(InvoiceItem.tid, InvoiceItem.username)
    q = q.group_by(InvoiceItem.tid, InvoiceItem.username)
    q = q.filter(
        InvoiceItem.paid == True,
        InvoiceItem.source == 'xml'
    )

    paid_transaction_ids = {i.tid: i.username for i in q}

    # Get the items matching the given invoice
    q = collection.for_invoice(invoice).query()
    q = q.with_entities(
        InvoiceItem.username,
        InvoiceItem.code,
        InvoiceItem.amount
    )
    q = q.filter(
        InvoiceItem.paid == False
    )
    q = q.order_by(
        InvoiceItem.username
    )

    # Sum up the items to virtual invoices
    invoices = []
    Invoice = namedtuple('Invoice', ('username', 'code', 'ref', 'amount'))

    def encode(code):
        try:
            return encode_invoice_code(code)
        except (RuntimeError, ValueError):
            return None

    for username, items in groupby(q, key=lambda i: i.username):
        amount = Decimal('0.00')

        for item in items:
            amount += item.amount

        invoices.append(Invoice(
            username=username,
            code=item.code,
            ref=encode(item.code),
            amount=amount
        ))

    # Hash the invoices by code (duplicates are technically possible)
    by_code = defaultdict(list)

    # hash the references as well
    by_ref = defaultdict(list)

    # Hash the invoices by amount (dpulicates are probable)
    by_amount = defaultdict(list)

    for invoice in invoices:
        by_code[invoice.code].append(invoice)
        by_amount[invoice.amount].append(invoice)
        by_ref[invoice.ref].append(invoice)

    # go through the transactions, comparing amount and code for a match
    transactions = tuple(extract_transactions(xml))

    codes = Counter(
        t.code for t in transactions
        if t.code and t.tid not in paid_transaction_ids
    )
    refs = Counter(
        t.reference for t in transactions
        if t.reference and t.tid not in paid_transaction_ids
    )

    for transaction in transactions:

        # credit transactions are completely irrelevant for us
        if not transaction.credit:
            continue

        if transaction.currency != currency:
            yield transaction
            continue

        if transaction.tid in paid_transaction_ids:
            transaction.paid = True
            transaction.username = paid_transaction_ids[transaction.tid]
            transaction.confidence = 1
            yield transaction
            continue

        # duplicate codes/references are marked as such, but not matched
        if codes[transaction.code] > 1 or refs[transaction.reference] > 1:
            transaction.duplicate = True

        if transaction.credit and transaction.currency == currency:
            code = transaction.code
            amnt = transaction.amount
            ref = transaction.reference

            code_usernames = {i.username for i in by_code.get(code, tuple())}
            amnt_usernames = {i.username for i in by_amount.get(amnt, tuple())}
            ref_usernames = {i.username for i in by_ref.get(ref, tuple())}

            combined = amnt_usernames & (code_usernames | ref_usernames)

            if len(combined) == 1:
                transaction.username = next(u for u in combined)
                transaction.confidence = 1.0
            elif len(code_usernames) == 1:
                transaction.username = next(u for u in code_usernames)
                transaction.confidence = 0.5
            elif len(ref_usernames) == 1:
                transaction.username = next(u for u in ref_usernames)
                transaction.confidence = 0.5
            elif len(amnt_usernames) == 1:
                transaction.username = next(u for u in amnt_usernames)
                transaction.confidence = 0.5

        yield transaction



from abc import ABCMeta, abstractmethod


class MatchableOccasion(metaclass=ABCMeta):
    """ Describes the interface required by the occasion class used by
    the algorithm.

    This allows us to untie our implementation from the database.

    """

    @property
    @abstractmethod
    def id(self):
        """ The id of the occasion. """

    @property
    @abstractmethod
    def max_spots(self):
        """ The maximum number of available spots. """

    @property
    @abstractmethod
    def exclude_from_overlap_check(self):
        """ True if bookings of this occasion are ignored during overlap
        checks.

        """


class MatchableBooking(metaclass=ABCMeta):
    """ Describes the interface required by the booking class used by
    the algorithm.

    This allows us to untie our implementation from the database.

    """

    def __eq__(self, other):
        """ The class must be comparable to other classes. """

    @abstractmethod
    def __hash__(self):
        """ The class must be hashable. """

    @property
    @abstractmethod
    def id(self):
        """ The id of the booking. """

    @property
    @abstractmethod
    def occasion_id(self):
        """ Returns the id of the occasion this booking belongs to. """

    @property
    @abstractmethod
    def attendee_id(self):
        """ Returns the id of the attendee this booking belongs to. """

    @property
    @abstractmethod
    def state(self):
        """ Returns the state of the booking, one of:

        * "open" (for unassigned bookings)
        * "accepted" (for already accepted bookings)
        * "blocked" (for bookings blocked by another accepted booking)

        """

    @property
    @abstractmethod
    def priority(self):
        """ Returns the priority of the booking. The higher the priority
        the further up the wishlist.

        Bookings further up the wishlist are first passed to the occasions.
        All things being equal (i.e. the scores of the other bookings), this
        leads to a higher chance of placement.

        """

    @property
    @abstractmethod
    def dates(self):
        """ Returns the dates of the booking. """



import sedate

from datetime import timedelta
from onegov.core.orm import Base
from onegov.core.orm.mixins import TimestampMixin
from onegov.core.orm.types import UUID
from psycopg2.extras import NumericRange
from sqlalchemy import Boolean
from sqlalchemy import column
from sqlalchemy import Column
from sqlalchemy import ForeignKey
from sqlalchemy import func
from sqlalchemy import Integer
from sqlalchemy import Numeric
from sqlalchemy import Text
from sqlalchemy.dialects.postgresql import ARRAY, INT4RANGE
from sqlalchemy.ext.hybrid import hybrid_property
from sqlalchemy.orm import relationship, object_session, validates
from sqlalchemy_utils import aggregated, observes
from uuid import uuid4


class Occasion(Base, TimestampMixin):
    """ Describes a single occurrence of an Activity. "Occurence" would have
    been a good word for it too, but that's used by onegov.event.

    So occasion it is.

    """

    __tablename__ = 'occasions'

    def __hash__(self):
        return hash(self.id)

    #: the public id of this occasion
    id = Column(UUID, primary_key=True, default=uuid4)

    #: Describes the meeting point of the occasion
    meeting_point = Column(Text, nullable=True)

    #: The expected age of participants
    age = Column(
        INT4RANGE, nullable=False, default=NumericRange(6, 17, bounds='[]'))

    #: The expected number of participants
    spots = Column(
        INT4RANGE, nullable=False, default=NumericRange(0, 10, bounds='[]'))

    #: A note about the occurrence
    note = Column(Text, nullable=True)

    #: The cost of the occasion (max value is 100'000.00), the currency is
    #: assumed to be CHF as this system will probably never be used outside
    #: Switzerland
    cost = Column(Numeric(precision=8, scale=2), nullable=True)

    #: The activity this occasion belongs to
    activity_id = Column(
        UUID, ForeignKey("activities.id", use_alter=True), nullable=False)

    #: The period this occasion belongs to
    period_id = Column(
        UUID, ForeignKey("periods.id", use_alter=True), nullable=False)

    #: True if the occasion has been cancelled
    cancelled = Column(Boolean, nullable=False, default=False)

    #: The durations defined by the associated dates
    durations = Column(Integer, default=0)

    #: The default order
    order = Column(Integer, default=0)

    #: Pretend like this occasion doesn't use any time
    exclude_from_overlap_check = Column(Boolean, nullable=False, default=False)

    #: Days on which this occasion is active
    active_days = Column(ARRAY(Integer), nullable=False, default=list)

    #: Weekdays on which this occasion is active
    weekdays = Column(ARRAY(Integer), nullable=False, default=list)

    @aggregated('period', Column(Boolean, default=False))
    def active(self):
        return column('active')

    @aggregated('accepted', Column(Integer, default=0))
    def attendee_count(self):
        return func.count('1')

    #: The bookings linked to this occasion
    bookings = relationship(
        'Booking',
        order_by='Booking.created',
        backref='occasion'
    )

    #: The dates associated with this occasion (loaded eagerly)
    dates = relationship(
        'OccasionDate',
        cascade='all,delete',
        order_by='OccasionDate.start',
        backref='occasion',
        lazy='joined',
    )

    accepted = relationship(
        'Booking',
        primaryjoin=("""and_(
            Booking.occasion_id == Occasion.id,
            Booking.state == 'accepted'
        )"""),
        viewonly=True
    )

    def on_date_change(self):
        """ Date changes are not properly propagated to the observer for
        some reason, so we do this manually with a hook.

        It's a bit of a hack, but multiple dates per occasion had to be
        added at the last minute..

        """
        self.observe_dates(self.dates)

    @observes('dates')
    def observe_dates(self, dates):
        if dates:
            self.durations = sum(set(d.duration for d in dates))
            self.order = int(min(d.start for d in dates).timestamp())
            self.active_days = [
                day
                for date in dates
                for day in date.active_days
            ]
            self.weekdays = list(set(
                day
                for date in dates
                for day in date.weekdays
            ))
        else:
            self.durations = 0
            self.order = -1
            self.active_days = []
            self.weekdays = []

    @validates('dates')
    def validate_dates(self, key, date):
        for o in self.dates:
            if o.id != date.id:
                assert not sedate.overlaps(
                    date.start, date.end, o.start, o.end)

        return date

    @hybrid_property
    def operable(self):
        return self.attendee_count >= self.spots.lower

    @hybrid_property
    def full(self):
        return self.attendee_count == self.spots.upper - 1

    @hybrid_property
    def available_spots(self):
        return self.spots.upper - 1 - self.attendee_count

    @property
    def max_spots(self):
        return self.spots.upper - 1

    def is_past_deadline(self, date):
        return date > self.deadline

    @property
    def deadline(self):
        """ The date until which this occasion may be booked (inclusive). """
        period = self.period

        if period.deadline_date is not None:
            return period.deadline_date

        min_date = min(d.start for d in self.dates)

        if period.deadline_days is not None:
            return (min_date - timedelta(days=period.deadline_days + 1)).date()

        return min_date.date()

    def cancel(self):
        from onegov.activity.collections import BookingCollection

        assert not self.cancelled
        period = self.period

        if not period.confirmed:
            def cancel(booking):
                booking.state = 'cancelled'
        else:
            bookings = BookingCollection(object_session(self))
            scoring = period.scoring

            def cancel(booking):
                bookings.cancel_booking(booking, scoring)

        for booking in self.bookings:
            assert booking.period_id == period.id
            cancel(booking)

        self.cancelled = True

    def is_permitted_birth_date(self, birth_date, wiggle_room=365):
        """ Returns true if the given birth date is considered to be in
        the age range of this occasion.

        The date inspected is the date of the first occurrence. In addition
        there's a wiggle room of x days which is applied to the exact age.

        By default attendees may be one year too young or too old.

        """

        if self.is_too_young(birth_date, wiggle_room):
            return False

        if self.is_too_old(birth_date, wiggle_room):
            return False

        return True

    def is_too_old(self, birth_date, wiggle_room=365):
        age = (self.dates[0].start.date() - birth_date).days
        max_age = (self.age.upper - 1) * 365.25 + wiggle_room

        return age > max_age

    def is_too_young(self, birth_date, wiggle_room=365):
        age = (self.dates[0].start.date() - birth_date).days
        min_age = self.age.lower * 365.25 - wiggle_room

        return age < min_age



from datetime import date
from onegov.activity.models.booking import Booking
from onegov.core.orm import Base
from onegov.core.orm.mixins import TimestampMixin
from onegov.core.orm.types import UUID
from onegov.search import ORMSearchable
from sqlalchemy import case, cast, func, select, and_, type_coerce
from sqlalchemy import Column
from sqlalchemy import Date
from sqlalchemy import Float
from sqlalchemy import ForeignKey
from sqlalchemy import Index
from sqlalchemy import Integer
from sqlalchemy import Numeric
from sqlalchemy import Text
from sqlalchemy.ext.hybrid import hybrid_property, hybrid_method
from sqlalchemy.orm import relationship, validates
from uuid import uuid4


class Attendee(Base, TimestampMixin, ORMSearchable):
    """ Attendees are linked to zero to many bookings. Each booking
    has an attendee.

    Though an attendee may be the same person as a username in reality, in the
    model we treat those subjects as separate.

    """

    __tablename__ = 'attendees'

    es_language = 'de'  # XXX add to database in the future
    es_properties = {
        'username': {'type': 'text'},
        'name': {'type': 'text'},
        'notes': {'type': 'localized'}
    }
    es_public = False

    @property
    def es_suggestion(self):
        return self.name

    def __hash__(self):
        return hash(self.id)

    #: the public id of the attendee
    id = Column(UUID, primary_key=True, default=uuid4)

    #: the user owning the attendee
    username = Column(Text, ForeignKey('users.username'), nullable=False)

    #: the name of the attendee (incl. first / lastname )
    name = Column(Text, nullable=False)

    #: birth date of the attendee for the age calculation
    birth_date = Column(Date, nullable=False)

    #: we use text for possible gender fluidity in the future ;)
    gender = Column(Text, nullable=True)

    #: notes about the attendee by the parents (e.g. allergies)
    notes = Column(Text, nullable=True)

    #: the maximum number of bookings the attendee wishes to get in each period
    limit = Column(Integer, nullable=True)

    #: access the user linked to this booking
    user = relationship('User')

    @validates('gender')
    def validate_gender(self, field, value):
        # for now we stay old-fashioned
        assert value in (None, 'male', 'female')
        return value

    @hybrid_property
    def age(self):
        today = date.today()
        birth = self.birth_date
        extra = (today.month, today.day) < (birth.month, birth.day) and 1 or 0

        return today.year - birth.year - extra

    @age.expression
    def age(self):
        return func.extract('year', func.age(self.birth_date))

    @hybrid_method
    def happiness(self, period_id):
        """ Returns the happiness of the attende in the given period.

        The happiness is a value between 0.0 and 1.0, indicating how many
        of the bookings on the wishlist were fulfilled.

        If all bookings were fulfilled, the happiness is 1.0, if no bookings
        were fulfilled the hapiness is 0.0.

        The priority of the bookings is taken into account. The decision on
        a high-priority booking has a higher impact than the decision on a
        low-priority booking. To model this we simply multiply the booking
        priority when summing up the happiness. So if a booking with priority
        1 is accepted, it is as if 2 bookings were accepted. If a booking
        with priority 1 is denied, it is as if 2 bookings were denied.

        """

        score = 0
        score_max = 0

        bookings = (b for b in self.bookings if b.period_id == period_id)

        for booking in bookings:
            score += booking.state == 'accepted' and booking.priority + 1
            score_max += booking.priority + 1

        # attendees without a booking have no known happiness (incidentally,
        # this works well with the sql expression below -> other default values
        # are harder to come by)
        if not score_max:
            return None

        return score / score_max

    @happiness.expression
    def happiness(cls, period_id):
        return select([
            # force the result to be a float instead of a decimal
            type_coerce(
                func.sum(
                    case([
                        (Booking.state == 'accepted', Booking.priority + 1),
                    ], else_=0)
                ) / cast(
                    # force the division to produce a float instead of an int
                    func.sum(Booking.priority) + func.count(Booking.id), Float
                ),
                Numeric(asdecimal=False)
            )
        ]).where(and_(
            Booking.period_id == period_id,
            Booking.attendee_id == cls.id
        )).label("happiness")

    #: The bookings linked to this attendee
    bookings = relationship(
        'Booking',
        order_by='Booking.created',
        backref='attendee'
    )

    __table_args__ = (
        Index('unique_child_name', 'username', 'name', unique=True),
    )



import sedate

from onegov.activity.models import Activity
from onegov.activity.utils import merge_ranges, overlaps
from onegov.core.collection import Pagination
from onegov.core.utils import increment_name
from onegov.core.utils import normalize_for_url
from sqlalchemy import or_, func, text, bindparam, Integer, select, column
from sqlalchemy.dialects.postgresql import array


class ActivityCollection(Pagination):

    def __init__(self, session, type='*', page=0,
                 tags=None,
                 states=None,
                 durations=None,
                 age_ranges=None,
                 owners=None,
                 period_ids=None,
                 dateranges=None,
                 weekdays=None):
        self.session = session
        self.type = type
        self.page = page
        self.tags = set(tags) if tags else set()
        self.states = set(states) if states else set()
        self.durations = set(durations) if durations else set()
        self.age_ranges = set(merge_ranges(age_ranges)) \
            if age_ranges else set()
        self.owners = set(owners) if owners else set()
        self.period_ids = set(period_ids) if period_ids else set()
        self.dateranges = set(dateranges) if dateranges else set()
        self.weekdays = set(weekdays) if weekdays else set()

    def __eq__(self, other):
        return self.type == other.type and self.page == other.page

    def subset(self):
        return self.query()

    @property
    def page_index(self):
        return self.page

    def page_by_index(self, index):
        return self.__class__(
            self.session,
            type=self.type,
            page=index,
            tags=self.tags,
            states=self.states,
            durations=self.durations,
            age_ranges=self.age_ranges,
            owners=self.owners,
            period_ids=self.period_ids,
            dateranges=self.dateranges,
            weekdays=self.weekdays
        )

    def contains_age_range(self, age_range):
        for r in self.age_ranges:
            if overlaps(r, age_range):
                return True
        return False

    @property
    def model_class(self):
        return Activity.get_polymorphic_class(self.type, Activity)

    def query_base(self):
        """ Returns the query based used by :meth:`query`. Overriding this
        function is useful to apply a general filter to the query before
        any other filter is applied.

        For example, a policy can be enforced that only allows public
        activites.

        """
        return self.session.query(self.model_class)

    def query(self):
        query = self.query_base()
        model_class = self.model_class

        if self.type != '*':
            query = query.filter(model_class.type == self.type)

        if self.tags:
            query = query.filter(model_class._tags.has_any(array(self.tags)))

        if self.states:
            query = query.filter(model_class.state.in_(self.states))

        if self.durations:
            conditions = tuple(
                func.coalesce(model_class.durations, 0).op('&')(int(d)) > 0
                for d in self.durations
            )
            query = query.filter(or_(*conditions))

        if self.age_ranges:
            conditions = []

            # SQLAlchemy needs unique parameter names if multiple text
            # queries with bind params are used
            for i, (min_age, max_age) in enumerate(self.age_ranges, start=1):
                stmt = "'[:min_{i},:max_{i}]'::int4range && ANY(ages)".format(
                    i=i)

                conditions.append(text(stmt).bindparams(
                    bindparam('min_' + str(i), min_age, type_=Integer),
                    bindparam('max_' + str(i), max_age, type_=Integer),
                ))

            query = query.filter(or_(*conditions))

        if self.owners:
            conditions = tuple(
                model_class.username == username for username in self.owners)

            query = query.filter(or_(*conditions))

        if self.period_ids:
            query = query.filter(
                model_class.period_ids.op('&&')(array(self.period_ids)))

        if self.dateranges:
            query = query.filter(
                model_class.active_days.op('&&')(array(
                    tuple(
                        dt.toordinal()
                        for start, end in self.dateranges
                        for dt in sedate.dtrange(start, end)
                    )
                ))
            )

        if self.weekdays:
            query = query.filter(
                model_class.weekdays.op('&&')(array(self.weekdays)))

        return query

    def for_filter(self,
                   tag=None,
                   state=None,
                   duration=None,
                   age_range=None,
                   owner=None,
                   period_id=None,
                   daterange=None,
                   weekday=None):
        """ Returns a new collection instance.

        The given tag is excluded if already in the list, included if not
        yet in the list. Same goes for the given state.

        Note that dateranges are excluded only if they match exactly. That is
        we don't care about overlaps at all. If the same exact daterange is
        found in the filter, it is excluded.

        """

        duration = int(duration) if duration is not None else None

        def toggle(collection, item):
            if item is None:
                return collection

            if item in collection:
                return collection - {item}
            else:
                return collection | {item}

        toggled = (
            toggle(collection, item) for collection, item in (
                (self.tags, tag),
                (self.states, state),
                (self.durations, duration),
                (self.age_ranges, age_range),
                (self.owners, owner),
                (self.period_ids, period_id),
                (self.dateranges, daterange),
                (self.weekdays, weekday)
            )
        )

        return self.__class__(self.session, self.type, 0, *toggled)

    def by_id(self, id):
        return self.query().filter(Activity.id == id).first()

    def by_name(self, name):
        return self.query().filter(Activity.name == name).first()

    def by_user(self, user):
        return self.query().filter(Activity.username == user.username)

    def by_username(self, username):
        return self.query().filter(Activity.username == username)

    @property
    def used_tags(self):
        """ Returns a list of all the tags used on *all* activites of
        the current type.

        """

        base = self.query_base().with_entities(
            func.skeys(self.model_class._tags).label('keys'))

        query = select([func.array_agg(column('keys'))], distinct=True)
        query = query.select_from(base.subquery())

        tags = self.session.execute(query).scalar()

        return set(tags) if tags else set()

    def get_unique_name(self, name):
        """ Given a desired name, finds a variant of that name that's not
        yet used. So if 'foobar' is already used, 'foobar-1' will be returned.

        """
        name = normalize_for_url(name)

        existing = self.session.query(self.model_class)
        existing = existing.filter(Activity.name.like('{}%'.format(name)))
        existing = existing.with_entities(Activity.name)
        existing = set(r[0] for r in existing.all())

        while name in existing:
            name = increment_name(name)

        return name

    def add(self, title, username, lead=None, text=None, tags=None, name=None):

        type = self.type != '*' and self.type or None

        name = name or self.get_unique_name(title)

        activity = self.model_class(
            name=name,
            title=title,
            tags=tags,
            type=type,
            username=username,
            lead=lead,
            text=text
        )

        self.session.add(activity)
        self.session.flush()

        return activity

    def delete(self, activity):
        for occasion in activity.occasions:
            self.session.delete(occasion)

        self.session.delete(activity)
        self.session.flush()



from datetime import datetime, timedelta
from onegov.activity import ActivityCollection
from onegov.activity import AttendeeCollection
from onegov.activity import BookingCollection
from onegov.activity import OccasionCollection
from onegov.activity import PeriodCollection
from onegov.core.utils import Bunch, module_path
from onegov.user import UserCollection
from pytest import fixture


@fixture(scope='function')
def owner(session):
    return UserCollection(session).add(
        username='owner@example.org',
        password='hunter2',
        role='editor'
    )


@fixture(scope='function')
def secondary_owner(session):
    return UserCollection(session).add(
        username='secondary@example.org',
        password='hunter2',
        role='editor'
    )


@fixture(scope='function')
def member(session):
    return UserCollection(session).add(
        username='member@example.org',
        password='hunter2',
        role='member'
    )


@fixture(scope='function')
def collections(session):
    return Bunch(
        activities=ActivityCollection(session),
        attendees=AttendeeCollection(session),
        bookings=BookingCollection(session),
        occasions=OccasionCollection(session),
        periods=PeriodCollection(session),
    )


@fixture(scope='function')
def prebooking_period(collections):
    """ Returns a period which is currently in the prebooking phase. """

    s, e = (
        datetime.utcnow() - timedelta(days=10),
        datetime.utcnow() + timedelta(days=10)
    )

    return collections.periods.add(
        title="Testperiod",
        prebooking=(s, e),
        execution=(e + timedelta(days=10), e + timedelta(days=20)),
        active=True
    )


@fixture(scope='function')
def execution_period(collections):
    """ Returns a period which is currently in the execution phase. """

    s, e = (
        datetime.utcnow() - timedelta(days=10),
        datetime.utcnow() + timedelta(days=10)
    )

    return collections.periods.add(
        title="Testperiod",
        prebooking=(s - timedelta(days=20), s - timedelta(days=10)),
        execution=(s, e),
        active=True
    )


@fixture(scope='function')
def inactive_period(collections):
    """ Returns a previously used period """

    s, e = (
        datetime.utcnow() - timedelta(days=100),
        datetime.utcnow() - timedelta(days=10)
    )

    return collections.periods.add(
        title="Testperiod",
        prebooking=(s - timedelta(days=20), s - timedelta(days=10)),
        execution=(s, e),
        active=False
    )


@fixture(scope='session')
def postfinance_xml():
    xml = 'camt.053_P_CH0309000000250090342_380000000_0_2016053100163801.xml'
    xml_path = module_path('onegov.activity', '/tests/fixtures/' + xml)

    with open(xml_path, 'r') as f:
        yield f.read()
